// mtest.C Version 7.43  Copyleft (c) 杉本典夫
// Time-stamp: <2025-03-08 16:29:19 tac>
//
// 層別クロス集計表を入力し、Mantel-Haenszelの検定(拡張法、一般化拡張法含む)と
// Breslow-Dayの検定を行う。
// 層別クロス集計表は計数値解析コマンド用の標準入力形式に従っている必要がある。
//
// 使用ライブラリ:libdans.so, libfun.so
// コンパイル法 :g++ -O -s -W mtest mtest.C
//               -I/usrlocal/include -L/usr/local/lib -ldans -lfun

#include <dans.H>               // dans用ヘッダファイル(mat.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "mtest  Version 7.43          Copyleft(Vo\\oV) 2025 杉本典夫",
  "書式:mtest  [<オプション> …]  [<クロス集計表ファイル> …]",
  "機能:層別クロス集計表に対してMantel-Haenszelの検定(拡張法含む)を行う。",
  "<オプション>",
  "  -cc[=<信頼係数>]:信頼係数を指定する。(0≦信頼係数<1)",
  "    -cc:-cc=0.95(オプション省略時)",
  "  -f=<ファイル名>:指定されたファイルからオプションを読み込む。",
  "  -h[=<コマンド名>]:オンラインマニュアルを表示する。(-h:-h=man)",
  "  -ngp={mhor|mhrd|mhrr|or|rr}[,…]:Ngraph用データを指定する。",
  "    mhor:調整オッズ比  mhrd:調整リスク差  mhrr:調整リスク比  or:オッズ比",
  "    rr:リスク比  オプション省略時:全て指定",
  "  -ngpname[=<出力ファイル基本名>]:Ngraph用データの出力ファイル基本名を指定。",
  "    オプション省略時:Ngraph用データ出力せず",
  "    -ngpname:-ngpname=mtest_ngp<2桁の連番>.dat",
  "  -res=[-][{def|all|bt|emt|gemt|lgt|m|mhor|mhrr|mt[{c|o}]|r|rrt|sts",
  "       |stt|tab[{fr|pc|pr|pt}]}][,…]:結果の出力項目を指定する。",
  "    -:-を付けた項目を出力しない(-res=-:全ての項目を出力しない)",
  "    def:オプション省略時と同じ項目を出力する  all:全ての項目を出力する",
  "    bt:Breslow-Dayの検定  emt:拡張Mantel検定  gemt:一般化拡張Mantel検定",
  "    lgt:ロジット検定",
  "    m:Bonferroni型多重検定  mhor:MHのオッズ比  mhrr:MHのリスク比",
  "    mt[{c|o}]:Mantel-Haenszel検定(c:連続修正有 o:連続修正無)  r:順位相関比",
  "    rrt:リスク比検定  sts:層の解析結果  stt:全体の解析結果",
  "    tab[{fr|pc|pr|pt}]:クロス集計表",
  "      fr:度数だけ出力  pr:行合計に対する％も出力  pc:列合計に対する％も出力",
  "      pt:全合計に対する％も出力",
  "    オプション省略時:データに合わせて自動判定",
  "  -skip[={-|+}]:不適な集計表をスキップするかどうかの指定。(-skip:-skip=+)"
};

const string Title = "=== 順序データの層別群間比較 ==="; // 表題
const string NGPNAMEdef = "mtest_ngp"; // 省略時Ngraphファイル基本名

class rci                       // リスク比と信頼区間クラス
{
private:
  int vn;                       // 例数
  double vr;                    // リスク比等
  double vrl;                   // リスク比下限
  double vru;                   // リスク比上限
public:
  rci() : vn(0), vr(0.0), vrl(0.0), vru(0.0) // デフォールトコンストラクタ
  {}
  rci(int n0, double r0, double rl0, double ru0) : vn(n0), vr(r0), vrl(rl0), vru(ru0)
  // コンストラクタ
  {}
  rci(const rci& rci0) : vn(rci0.vn), vr(rci0.vr), vrl(rci0.vrl), vru(rci0.vru)
  // コピーコンストラクタ
  {}
  virtual ~rci()                // デストラクタ
  {}
  const rci& operator=(const rci& rci0); // コピー代入演算子
  int n() const                          // 例数を返す
  { return vn; }
  int n(int n0)                 // 例数を設定する
  { return (vn = n0); }
  double r() const              // リスク比を返す
  { return vr; }
  double r(double r0)           // リスク比を設定する
  { return (vr = r0); }
  double rl() const             // リスク比下限を返す
  { return vrl; }
  double rl(double rl0)         // リスク比下限を設定する
  { return (vrl = rl0); }
  double ru() const             // リスク比上限を返す
  { return vru; }
  double ru(double ru0)         // リスク比上限を設定する
  { return (vru = ru0); }
};

const rci& rci::operator=(const rci& rci0) // コピー代入演算子
{
  if (&rci0 == this) return *this; // 自分自身への代入を防ぐ
  vn = rci0.vn;  vr = rci0.vr;  vrl = rci0.vrl;  vru = rci0.vru;
  return *this;                 // 自分自身への参照を返す
}

using vecrci = vector<rci>;    // リスク比と信頼区間ベクトルの短名定義

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  testtype vmt;                 // Mantel-Haenszel検定の種類
  string vngpname;              // Ngraphファイル基本名
  string vngpnamen;             // Ngraphファイル名(番号付き)
  lstr vngplst;                 // Ngraph出力リスト
  mapsb vngpmap;                // Ngraph出力の有無マップ
  lstr vreslst;                 // 分析結果出力リスト
  mapsb vresmap;                // 分析結果出力の有無マップ
  bool vskip;                   // 不適集計表スキップフラッグ
  tabtype vtab;                 // 度数集計費用の種類
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  testtype mt() const           // Mantel-Haenszel検定の種類を返す
  { return vmt; }
  const string& ngpname(int fno); // Ngraphデータファイル名を返す
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  bool res(const string& code); // 解析結果出力フラッグを返す
  bool skip() const             // 不適集計表スキップフラッグを返す
  { return vskip; }
  tabtype tab() const           // 度数集計表の種類を返す
  { return vtab; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check();                  // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vngpname(), vngpnamen(), vngplst(), vngpmap(), vmt(testcor), vreslst(), vresmap(), vskip(false), vtab(tabfprr)
// コンストラクタ
{
  if (c_pname == 0) return;     // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

const string& option::ngpname(int fno) // Ngraphデータファイル名を返す
{
  if (vngpname == "") return vngpname; // -ngpname指定なし
  ostringstream ostr;
  if (fno < 10) ostr << vngpname << "0" << fno << ".dat";
  else ostr << vngpname << fno << ".dat";
  vngpnamen = (string)(ostr.str());
  return vngpnamen;             // 番号付きファイル名を返す
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  mapsb_citr ngpitr = vngpmap.find(code);
  if (ngpitr == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  mapsb_citr resitr = vresmap.find(code);
  if (resitr == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "cc")           // 信頼係数の指定
  {
    vcc = opt[1].empty() ? CCdef : strtodf(opt[1]);
    if (vcc < 0.0 || 1.0 <= vcc) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル基本名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
  }
  else if (opt[0] == "res")     // 解析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "skip")   // 不適集計表スキップ指定
  {
    if (opt[1].empty() || opt[1] == "+") vskip = true;
    else if (opt[1] == "-") vskip = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("ヘルプ表示");
  if (err < 0) throw option_error(string("オプション指定違反→") + optstr);
}

int option::check()             // オプションのチェックとセット
{
  if (set_ngp() < 0) return -2; // -ngpオプションのセット
  if (set_res() < 0) return -3; // -resオプションのセット
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["mhor"] = vngpmap["mhrd"] = vngpmap["mhrr"] = vngpmap["or"] = vngpmap["rr"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "-ngpnameオプションの指定が必要！");
    vngpmap["mhor"] = vngpmap["mhrd"] = vngpmap["mhrr"] = vngpmap["or"] = vngpmap["rr"] = false;
    for (const string& ngpstr : vngplst)
    {
      mapsb_citr ngpitr = vngpmap.find(ngpstr);
      if (ngpitr == vngpmap.end())
      { return disperr(vpname, "-ngp=" + ngpstr + "オプション指定が不適当！"); }
      else
      { vngpmap[ngpstr] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["mt"] = vresmap["sts"] = vresmap["stt"] = vresmap["tab"] = true;
  vresmap["bt"] = vresmap["emt"] = vresmap["gemt"] = vresmap["lgt"] = vresmap["m"] = vresmap["mhor"] = vresmap["mhrr"] = vresmap["r"] = vresmap["rrt"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["mt"] = vresmap["sts"] = vresmap["stt"] = vresmap["tab"] = false;
    for (const string& resstr : vreslst)
    {
      mapsb_citr resitr = vresmap.find(resstr);
      if (resstr == "-")        // 全て出力せず
      { vresmap["res"] = false; }
      else if (resstr[0] == '-' && resstr.size() > 1) // 特定の項目を出力せず
      {
        resitr = vresmap.find(resstr.substr(1));
        if (resitr == vresmap.end())
        { return disperr(vpname, "-res=" + resstr + "オプション指定が不適当！"); }
        else
        { vresmap[resstr.substr(1)] = false; }
      }
      else if (resstr == "def") // 省略時出力
      {
        vresmap["def"] = vresmap["mt"] = vresmap["sts"] = vresmap["stt"] = vresmap["tab"] = true;
      }
      else if (resstr == "all") // 全て出力
      {
        vresmap["def"] = vresmap["mt"] = vresmap["sts"] = vresmap["stt"] = vresmap["tab"] = vresmap["bt"] = vresmap["emt"] = vresmap["gemt"] = vresmap["lgt"] = vresmap["m"] = vresmap["mhor"] = vresmap["mhrr"] = vresmap["r"] = vresmap["rrt"] = true;
        vmt = testcor;  vtab = tabfprr;
      }
      else if (resstr == "mtc")    // 連続修正有
      { vresmap["mt"] = true;  vmt = testcor; }
      else if (resstr == "mto")    // 連続修正無
      { vresmap["mt"] = true;  vmt = testorg; }
      else if (resstr == "tabfr")  // 度数だけ出力
      { vresmap["tab"] = true;  vtab = tabfreq; }
      else if (resstr == "tabpc")  // 列合計に対する％も出力
      { vresmap["tab"] = true;  vtab = tabfprc; }
      else if (resstr == "tabpr")  // 行合計に対する％も出力
      { vresmap["tab"] = true;  vtab = tabfprr; }
      else if (resstr == "tabpt")  // 全体合計に対する％も出力
      { vresmap["tab"] = true;  vtab = tabfprt; }
      else if (resitr == vresmap.end())
      { return disperr(vpname, "-res=" + resstr + "オプション指定が不適当！"); }
      else
      { vresmap[resstr] = true; }
    }
  }
  return 0;                     // エラーなし
}

class varclass                  // 変数クラス
{
private:
  int vnftabok;                 // 有効な度数表の数
  int vnt;                      // 全体の合計例数
  vecrci vmhorvec;              // Mantel-Haenzelの調整オッズ比
  vecrci vmhrdvec;              // Mantel-Haenzelの調整リスク差
  vecrci vmhrrvec;              // Mantel-Haenzelの調整リスク比
  vecrci vorvec;                // オッズ比
  vecrci vrrvec;                // リスク比
  void cal_mhstat(option& opt, listidd& idlst, const listftab& ftablst);
  // Mantel-Haenzelの各種統計量を計算する
  void cal_mtest(option& opt, int fidx, const listftab& ftablst0, int nt);
  // Mantel-Haenszelの検定を行い各種統計量を返す
  void cal_rrtest(option& opt, int fidx, const listftab& ftablst0, int nt);
  // リスク比検定を行い各種統計量を返す
  void cal_lgtest(option& opt, int fidx, const listftab& ftablst0, int nt);
  // ロジット検定(オッズ比検定)を行い各種統計量を返す
public:
  varclass() : vnftabok(0), vmhorvec(), vmhrdvec(), vmhrrvec(), vorvec(), vrrvec()
  // デフォールトコンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  int nftabok() const           // 有効な度数表の数を返す
  { return vnftabok; }
  int nt() const                // 全体の合計例数を返す
  { return vnt; }
  const vecrci& mhorvec() const // Mantel-Haenzelの調整オッズ比を返す
  { return vmhorvec; }
  const vecrci& mhrdvec() const // Mantel-Haenzelの調整リスク差を返す
  { return vmhrdvec; }
  const vecrci& mhrrvec() const // Mantel-Haenzelの調整リスク比を返す
  { return vmhrrvec; }
  const vecrci& orvec() const   // オッズ比を返す
  { return vorvec; }
  const vecrci& rrvec() const   // リスク比を返す
  { return vrrvec; }
  void cal(option& opt, listidd& idlst, const listftab& ftablst);
  // 各種の計算を行い変数を設定する
};

void varclass::cal_mhstat(option& opt, listidd& idlst, const listftab& ftablst)
// Mantel-Haenzelの各種統計量を計算する
{
  listidd::iterator iitr = idlst.begin();
  listftab::const_iterator fitr = ftablst.begin();
  size_t nr = fitr->nr(), nc = fitr->nc();
  ftable ftabt(*fitr);          // 合計度数表用
  ftabt.fclear();
  int fidx = 0;
  vnt = 0;
  while (fitr != ftablst.end())
  {
    ftable ftabtrm(*fitr);
    ftabtrm.trm();
    listftab ftablsti;
    ftablsti.push_back(ftabtrm);
    if (fitr->nr() == nr && fitr->nc() == nc && ftabtrm.nr() > 1 && ftabtrm.nc() > 1)
    {
      cal_mtest(opt, fidx, ftablsti, ftabtrm.ftotal());
      cal_rrtest(opt, fidx, ftablsti, ftabtrm.ftotal());
      cal_lgtest(opt, fidx, ftablsti, ftabtrm.ftotal());
      vnt += fitr->ftotal();
      for (size_t i = 1; i <= nr; ++i)
      {
        for (size_t j = 1; j <= nc; ++j) ftabt.f(i, j, ftabt.f(i, j) + fitr->f(i, j));
      }
      ++fidx;
    }
    ++fitr;  ++iitr;
  }
  ftabt.set();
  listftab ftablstt;
  ftablstt.push_back(ftabt);
  cal_mtest(opt, fidx, ftablstt, vnt);
  cal_rrtest(opt, fidx, ftablstt, vnt);
  cal_lgtest(opt, fidx, ftablstt, vnt);
  cal_mtest(opt, fidx + 1, ftablst, vnt);
  cal_rrtest(opt, fidx + 1, ftablst, vnt);
  cal_lgtest(opt, fidx + 1, ftablst, vnt);
}

void varclass::cal_mtest(option& opt, int fidx, const listftab& ftablst0, int nt)
// Mantel-Haenszelの検定を行い各種統計量を計算する
{
  double rdm = 0.0, vrdm = 0.0, rrm = 0.0, rra = 0.0, rrb = 0.0, rrc = 0.0;
  double orm = 0.0, ora = 0.0, orb = 0.0, orc = 0.0, cmm = 0.0, cmr = 0.0;
  int dfm = 1;
  int dfr = mtest(rdm, vrdm, rrm, rra, rrb, rrc, orm, ora, orb, orc, cmm, cmr, ftablst0, opt.mt());
  double ln_rrm = (rrm == 0.0) ? 0.0 : log(rrm);
  double ln_orm = (orm == 0.0) ? 0.0 : log(orm);
  double z = ptoz(1.0 - opt.cc());
  double ci = z * sqrt(vrdm);
  double cil = rdm - ci, ciu = rdm + ci;
  vmhrdvec[fidx].n(nt);  vmhrdvec[fidx].r(rdm);
  vmhrdvec[fidx].rl(cil);  vmhrdvec[fidx].ru(ciu);
  if (0.0 < rrm)
  {
    double rrt = 2.0 * rra * rrb + z * z * rrc;
    double vrrt = (4.0 * rra * rrb + z * z * rrc) * rrc;
    if (rrb < FDeps)
    { cil = ciu = rrt; }
    else
    {
      cil = (rrt - z * sqrt(vrrt)) / 2.0 / rrb / rrb;
      ciu = (rrt + z * sqrt(vrrt)) / 2.0 / rrb / rrb;
    }
  }
  vmhrrvec[fidx].n(nt);  vmhrrvec[fidx].r(rrm);
  vmhrrvec[fidx].rl(cil);  vmhrrvec[fidx].ru(ciu);
  if (0.0 < orm)
  {
    double ort = 2.0 * ora * orb;
    double vort = (4.0 * ora * orb + z * z * orc) * orc;
    if (orb < FDeps)
    { cil = ciu = orb; }
    else
    {
      cil = (ort + z * z * orc - z * sqrt(vort)) / 2.0 / orb / orb;
      ciu = (ort - z * z * orc + z * sqrt(vort)) / 2.0 / orb / orb;
    }
  }
  vmhorvec[fidx].n(nt);  vmhorvec[fidx].r(orm);
  vmhorvec[fidx].rl(cil);  vmhorvec[fidx].ru(ciu);
}

void varclass::cal_rrtest(option& opt, int fidx, const listftab& ftablst0, int nt)
// リスク比検定を行い各種統計量を計算する
{
  double rr = 0.0, vrr = 0.0, ca = 0.0, ch = 0.0;
  int nc = rrtest(rr, vrr, ca, ch, ftablst0);
  int dfa = 1, dfh = nc - 1;
  if (nc < 1) return;
  ldbl pr = ctop(ca, dfa);
  double ln_rr = (rr == 0.0) ? 0.0 : log(rr);
  double z = ptoz(1.0 - opt.cc());
  double ci = z * sqrt(vrr);
  double ln_rrl = ln_rr - ci, ln_rru = ln_rr + ci;
  double rrl = exp(ln_rrl), rru = exp(ln_rru);
  vrrvec[fidx].n(nt);  vrrvec[fidx].r(rr);
  vrrvec[fidx].rl(rrl);  vrrvec[fidx].ru(rru);
}

void varclass::cal_lgtest(option& opt, int fidx, const listftab& ftablst0, int nt)
// ロジット検定(オッズ比検定)を行い各種統計量を計算する
{
  double odr = 0.0, vodr = 0.0, ca = 0.0, ch = 0.0;
  int dfa = 1, dfh = lgtest(odr, vodr, ca, ch, ftablst0);
  ldbl pr = ctop(ca, dfa);
  double ln_odr = (odr == 0.0) ? 0.0 : log(odr);
  double z = ptoz(1.0 - opt.cc());
  double ci = z * sqrt(vodr);
  double ln_odrl = ln_odr - ci, ln_odru = ln_odr + ci;
  double odrl = exp(ln_odrl), odru = exp(ln_odru);
  vorvec[fidx].n(nt);  vorvec[fidx].r(odr);
  vorvec[fidx].rl(odrl);  vorvec[fidx].ru(odru);
}

void varclass::cal(option& opt, listidd& idlst, const listftab& ftablst)
// 各種の計算を行い変数を設定する
{
  vnftabok = 0;
  if (ftablst.empty()) return;  // 度数表なし
  listftab::const_iterator fitr = ftablst.begin();
  size_t nr = fitr->nr(), nc = fitr->nc();
  while (fitr != ftablst.end())
  {
    ftable ftabtrm(*fitr);
    ftabtrm.trm();
    if (fitr->nr() == nr && fitr->nc() == nc && ftabtrm.nr() > 1 && ftabtrm.nc() > 1) ++vnftabok;
    ++fitr;
  }
  vmhorvec.resize(vnftabok + 2);
  vmhrdvec.resize(vnftabok + 2);
  vmhrrvec.resize(vnftabok + 2);
  vorvec.resize(vnftabok + 2);
  vrrvec.resize(vnftabok + 2);
  cal_mhstat(opt, idlst, ftablst);
}

class mtestclass                // mtestメインクラス
{
private:
  string fname;                 // ファイル名
  void put_ftable(ostream& fout, option& opt, listidd& idlst, const listftab& ftablst, const varclass& var) const;
  // 度数表と解析結果を出力する
  void put_resid(ostream& fout, int idno, idd& id) const;
  // 解析結果のタイトルとデータ識別情報を出力する
  void put_ftabi(ostream& fout, option& opt, int idno, const ftable& ftabi, int nftab, bool ftabt) const;
  // 個々の度数集表と検定結果を出力する
  void put_overall(ostream& fout, option& opt, const listftab& ftablst, const varclass& var) const;
  // 全体の解析結果を出力する
  void put_mtest(ostream& fout, option& opt, int fidx, const listftab& ftablst0, int nftab, int nt, bool ftabt) const;
  // Mantel-Haenzelの検定結果を出力する
  void put_mr(ostream& fout, const listftab& ftablst0, int nt) const;
  // Mantel-Haenszelの順位相関比を出力する
  void put_lgtest(ostream& fout, option& opt, int fidx, const listftab& ftablst0, int nftab, int nt, bool ftabt) const;
  // ロジット検定(オッズ比検定)の結果を出力する
  void put_rrtest(ostream& fout, option& opt, int fidx, const listftab& ftablst0, int nftab, int nt, bool ftabt) const;
  // リスク比検定の結果を出力する
  void put_emtest(ostream& fout, option& opt, const listftab& ftablst0, int nftab, int nt, bool ftabt) const;
  // 拡張Mantel検定の結果を出力する
  void put_gemtest(ostream& fout, option& opt, const listftab& ftablst0, int nftab, int nt, bool ftabt) const;
  // 一般化拡張Mantel検定の結果を出力する
  void plot_ngp(int fno, option& opt, listidd& idlst, const varclass& var);
  // Ngraph用データを出力する
  void plot_ngpid(ofstream& fout, listidd& idlst) const;
  // Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_rcivec(ofstream& fout, const string& rname, int nftab, int nt, const vecrci& rcivec) const;
  // 個々のリスク比と信頼区間をNgraph用に出力する
public:
  mtestclass() : fname()        // コンストラクタ
  {}
  ~mtestclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt, listidd& idlst, listftab& ftablst);
  // 度数集計表を読み込む
  void put(option& opt, listidd& idlst, listftab& ftablst, int fno);
  // 集計結果を出力する
};

void mtestclass::put_ftable(ostream& fout, option& opt, listidd& idlst, const listftab& ftablst, const varclass& var) const
// 度数表と解析結果を出力する
{
  listidd::iterator iitr = idlst.begin();
  listftab::const_iterator fitr = ftablst.begin();
  size_t nr = fitr->nr(), nc = fitr->nc();
  idd idt(*iitr);               // 合計度数表用
  ftable ftabt(*fitr);          // 合計度数表用
  ftabt.fclear();
  int nt = 0, idno = 1;
  bool puttab = false;
  while (fitr != ftablst.end())
  {
    if (opt.res("sts"))         // 層ごとの結果出力
    {
      if (puttab) fout << "\n\n";
      put_resid(fout, idno, *iitr);
      put_ftabi(fout, opt, idno, *fitr, var.nftabok(), false);
      puttab = true;
    }
    ftable ftabtrm(*fitr);
    ftabtrm.trm();
    if (fitr->nr() != nr || fitr->nc() != nc || ftabtrm.nr() < 2 || ftabtrm.nc() < 2)
    {
      if (opt.res("sts")) fout << "※不適切な度数表のため全体解析から除外\n";
    }
    else
    {
      nt += fitr->ftotal();
      for (size_t i = 1; i <= nr; ++i)
      {
        for (size_t j = 1; j <= nc; ++j) ftabt.f(i, j, ftabt.f(i, j) + fitr->f(i, j));
      }
    }
    ++fitr;  ++iitr;  ++idno;
  }
  if (var.nftabok() < 2 || opt.res("stt") == false) return; // 層だけ出力
  ftabt.set();
  if (puttab) fout << "\n\n";
  put_resid(fout, 0, idt);
  put_ftabi(fout, opt, idno, ftabt, var.nftabok(), true);
}

void mtestclass::put_resid(ostream& fout, int idno, idd& id) const
// 解析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  if (idno > 0) id.put_stra(fout);
  else fout << "層別内容:全度数表合計\n\n";
  for (size_t i = 0; i < id.ni(); ++i)
  {
    switch (id.eitem(i).aval())
    {
    case avalrow : fout << "群項目(縦)  :" << id.item(i);
      if (idno == 0) fout << "-合計";
      fout << "\n";
      break;
    case avalcol : fout << "順序項目(横):" << id.item(i);
      if (idno == 0) fout << "-合計";
      fout << "\n";
      break;
    default : id.put_item(fout, i);
    }
  }
}

void mtestclass::put_ftabi(ostream& fout, option& opt, int idno, const ftable& ftabi, int nftab, bool ftabt) const
// 個々の度数表と解析結果を出力する
{
  bool puttab = true;           // put_resid()を出力した後で実行するため
  if (opt.res("tab")) putftable(fout, ftabi, opt.tab());
  ftable ftabtrm(ftabi);
  ftabtrm.trm();
  if (ftabtrm.nr() < 2 || ftabtrm.nc() < 2) return; // 検定不可能
  listftab ftablsti;
  ftablsti.push_back(ftabtrm);
  if (ftabtrm.nr() == 2)        // 2群
  {
    if (ftabtrm.nc() == 2)      // 2順位
    {
      if (opt.res("mt"))
      {
        if (puttab) fout << "\n";
        put_mtest(fout, opt, idno - 1, ftablsti, nftab, ftabtrm.ftotal(), ftabt);
        puttab = true;
      }
      if (opt.res("rrt"))
      {
        if (puttab) fout << "\n";
        put_rrtest(fout, opt, idno - 1, ftablsti, nftab, ftabtrm.ftotal(), ftabt);
        puttab = true;
      }
      if (opt.res("lgt"))
      {
        if (puttab) fout << "\n";
        put_lgtest(fout, opt, idno - 1, ftablsti, nftab, ftabtrm.ftotal(), ftabt);
        puttab = true;
      }
      if (opt.res("emt"))
      {
        if (puttab) fout << "\n";
        put_emtest(fout, opt, ftablsti, nftab, ftabtrm.ftotal(), ftabt);
        puttab = true;
      }
    }
    else                        // 3順位以上(省略時は拡張Mantel検定出力)
    {
      if (opt.res("def") || opt.res("emt"))
      {
        if (puttab) fout << "\n";
        put_emtest(fout, opt, ftablsti, nftab, ftabtrm.ftotal(), ftabt);
        puttab = true;
      }
    }
  }
  if ((ftabi.nr() > 2 && opt.res("def")) || opt.res("gemt"))
  {                       // 3群以上(省略時は一般化拡張Mantel検定出力)
    if (puttab) fout << "\n";
    put_gemtest(fout, opt, ftablsti, nftab, ftabtrm.ftotal(), ftabt);
  }
}

void mtestclass::put_overall(ostream& fout, option& opt, const listftab& ftablst, const varclass& var) const
// 全体の解析結果を出力する
{
  fout << "\n○全体\n";
  bool puttab = false;
  listftab::const_iterator fitr = ftablst.begin();
  if (fitr->nr() == 2)          // 2群
  {
    if (fitr->nc() == 2)        // 2順位
    {
      if (opt.res("mt"))
      {
        put_mtest(fout, opt, var.nftabok() + 1, ftablst, var.nftabok(), var.nt(), false);
        puttab = true;
      }
      if (opt.res("rrt"))
      {
        if (puttab) fout << "\n";
        put_rrtest(fout, opt, var.nftabok() + 1, ftablst, var.nftabok(), var.nt(), false);
        puttab = true;
      }
      if (opt.res("lgt"))
      {
        if (puttab) fout << "\n";
        put_lgtest(fout, opt, var.nftabok() + 1, ftablst, var.nftabok(), var.nt(), false);
        puttab = true;
      }
      if (opt.res("emt"))
      {
        if (puttab) fout << "\n";
        puttab = true;
        put_emtest(fout, opt, ftablst, var.nftabok(), var.nt(), false);
      }
    }
    else                        // 3順位以上(省略時は拡張Mantel検定出力)
    {
      if (opt.res("def") || opt.res("emt"))
      {
        put_emtest(fout, opt, ftablst, var.nftabok(), var.nt(), false);
        puttab = true;
      }
    }
  }
  if ((fitr->nr() > 2 && opt.res("def")) || opt.res("gemt"))
  {                       // 3群以上(省略時は一般化拡張Mantel検定出力)
    if (puttab) fout << "\n";
    put_gemtest(fout, opt, ftablst, var.nftabok(), var.nt(), false);
  }
}

void mtestclass::put_mtest(ostream& fout, option& opt, int fidx, const listftab& ftablst0, int nftab, int nt, bool ftabt) const
// Mantel-Haenszelの検定結果を出力する
{
  double rdm = 0.0, vrdm = 0.0, rrm = 0.0, rra = 0.0, rrb = 0.0, rrc = 0.0;
  double orm = 0.0, ora = 0.0, orb = 0.0, orc = 0.0, cmm = 0.0, cmr = 0.0;
  int dfm = 1;
  int dfr = mtest(rdm, vrdm, rrm, rra, rrb, rrc, orm, ora, orb, orc, cmm, cmr, ftablst0, opt.mt());
  double ln_rrm = (rrm == 0.0) ? 0.0 : log(rrm);
  double ln_orm = (orm == 0.0) ? 0.0 : log(orm);
  double z = ptoz(1.0 - opt.cc());
  double ci = z * sqrt(vrdm);
  double cil = rdm - ci, ciu = rdm + ci;
  ldbl pr = 0.0L;
  fout << "・Mantel-Haenszel検定\n";
  if (opt.res("r")) put_mr(fout, ftablst0, nt);
  fout << "  ";
  if (dfr > 0) fout << "調整";
  fout << "リスク差(risk difference):RD=" << rdm << "\n";
  fout << "    " << opt.cc() * 100.0 << "%信頼区間:下限=" << cil << "  上限=" << ciu << "\n";
  if (0.0 < rrm)
  {
    double rrt = 2.0 * rra * rrb + z * z * rrc;
    double vrrt = (4.0 * rra * rrb + z * z * rrc) * rrc;
    if (rrb < FDeps)
    { cil = ciu = rrt; }
    else
    {
      cil = (rrt - z * sqrt(vrrt)) / 2.0 / rrb / rrb;
      ciu = (rrt + z * sqrt(vrrt)) / 2.0 / rrb / rrb;
    }
  }
  if (opt.res("mhrr"))          // リスク比出力
  {
    fout << "  ";
    if (dfr > 0) fout << "調整";
    fout << "リスク比(risk ratio):RR=" << rrm << "  ln(RR)=" << ln_rrm << "\n";
    if (0.0 < rrm) fout << "    " << opt.cc() * 100.0 << "%信頼区間:下限=" << cil << "  上限=" << ciu << "\n";
  }
  if (0.0 < orm)
  {
    double ort = 2.0 * ora * orb;
    double vort = (4.0 * ora * orb + z * z * orc) * orc;
    if (orb < FDeps)
    { cil = ciu = orb; }
    else
    {
      cil = (ort + z * z * orc - z * sqrt(vort)) / 2.0 / orb / orb;
      ciu = (ort - z * z * orc + z * sqrt(vort)) / 2.0 / orb / orb;
    }
  }
  if (opt.res("mhor"))          // オッズ比出力
  {
    fout << "  ";
    if (dfr > 0) fout << "調整";
    fout << "オッズ比(odds ratio):OR=" << orm << "  ln(OR)=" << ln_orm << "\n";
    if (0.0 < orm) fout << "    " << opt.cc() * 100.0 << "%信頼区間:下限=" << cil << "  上限=" << ciu << "\n";
  }
  pr = ctop(cmm, dfm);
  fout << "  共通性の検定";
  if (opt.mt() == testorg) fout << "(修正無)";
  else fout << "(修正有)";
  fout << ":χ2乗=" << cmm << "  自由度=" << dfm << "  p=" << pr << pind(pr) << "\n";
  if (ftabt) return;            // 全度数表合計
  if (dfr < 1)
  {
    if (opt.res("m") == false) return;
    pr *= (ldbl)nftab;
    if (pr > 1.0L) pr = 1.0L;
    fout << "  Bonferroni型多重検定:p=" << pr << pind(pr) << "\n";
    z = ptoz((1.0 - opt.cc()) / (double)nftab);
    ci = z * sqrt(vrdm);
    fout << "    リスク差の" << opt.cc() * 100.0 << "%同時信頼区間:下限=" << rdm - ci << "  上限=" << rdm + ci << endl;
    return;
  }
  pr = ctop(cmr, dfr);
  fout << "  異質性の検定:χ2乗=" << cmr << "  自由度=" << dfr << "  p=" << pr << pind(pr) << "\n";
  if (opt.res("bt"))
  {
    dfr = mbtest(orm, cmm, cmr, ftablst0, opt.mt());
    pr = ctop(cmr, dfr);
    fout << "  Breslow-Dayの検定(異質性):χ2乗=" << cmr << "  自由度=" << dfr << "  p=" << pr << pind(pr) << "\n";
  }
}

void mtestclass::put_mr(ostream& fout, const listftab& ftablst0, int nt) const
// Mantel-Haenszelの順位相関比を出力する
{
  double rdm = 0.0, vrdm = 0.0, rrm = 0.0, rra = 0.0, rrb = 0.0, rrc = 0.0;
  double orm = 0.0, ora = 0.0, orb = 0.0, orc = 0.0, cmm = 0.0, cmr = 0.0;
  mtest(rdm, vrdm, rrm, rra, rrb, rrc, orm, ora, orb, orc, cmm, cmr, ftablst0, testorg);
  double r2 = cmm / (double)(nt - 1);
  int dfm = 1;
  ldbl pr = ctop(cmm, dfm);
  fout << "  順位相関比:η=" << sqrt(r2) << "  寄与率:η2乗=" << r2 << "\n";
  fout << "  順位相関比の検定:χ2乗=" << cmm << "  自由度=" << dfm << "  p=" << pr << pind(pr) << endl;
}

void mtestclass::put_rrtest(ostream& fout, option& opt, int fidx, const listftab& ftablst0, int nftab, int nt, bool ftabt) const
// リスク比検定の検定結果を出力する
{
  double rr = 0.0, vrr = 0.0, ca = 0.0, ch = 0.0;
  int nc = rrtest(rr, vrr, ca, ch, ftablst0);
  int dfa = 1, dfh = nc - 1;
  if (nc < 1)
  {
    fout << "\n・リスク比検定:不可能\n";
    return;
  }
  ldbl pr = ctop(ca, dfa);
  double ln_rr = (rr == 0.0) ? 0.0 : log(rr);
  double z = ptoz(1.0 - opt.cc());
  double ci = z * sqrt(vrr);
  double ln_rrl = ln_rr - ci, ln_rru = ln_rr + ci;
  double rrl = exp(ln_rrl), rru = exp(ln_rru);
  fout << "・リスク比検定\n";
  fout << "  ";
  if (dfh > 0) fout << "調整";
  fout << "リスク比(risk ratio):RR=" << rr << "  ln(RR)=" << ln_rr << "\n";
  fout << "    " << opt.cc() * 100.0 << "%信頼区間:下限=" << rrl << "  上限=" << rru << "\n";
  fout << "  ";
  if (dfh > 0) fout << "調整";
  fout << "リスク比の検定:χ2乗=" << ca << "  自由度=" << dfa << "  p=" << pr << pind(pr) << "\n";
  if (ftabt) return;            // 全度数表合計
  if (nc > 1)
  {
    pr = ctop(ch, dfh);
    fout << "  異質性の検定:χ2乗=" << ch << "  自由度=" << dfh << "  p=" << pr << pind(pr) << endl;
  }
}

void mtestclass::put_lgtest(ostream& fout, option& opt, int fidx, const listftab& ftablst0, int nftab, int nt, bool ftabt) const
// ロジット検定(オッズ比検定)の検定結果を出力する
{
  double odr = 0.0, vodr = 0.0, ca = 0.0, ch = 0.0;
  int dfa = 1, dfh = lgtest(odr, vodr, ca, ch, ftablst0);
  ldbl pr = ctop(ca, dfa);
  double ln_odr = (odr == 0.0) ? 0.0 : log(odr);
  double z = ptoz(1.0 - opt.cc());
  double ci = z * sqrt(vodr);
  double ln_odrl = ln_odr - ci, ln_odru = ln_odr + ci;
  double odrl = exp(ln_odrl), odru = exp(ln_odru);
  fout << "・ロジット検定\n";
  fout << "  ";
  if (dfh > 0) fout << "調整";
  fout << "オッズ比(odds ratio):OR=" << odr << "  ln(OR)=" << ln_odr << "\n";
  fout << "    " << opt.cc() * 100.0 << "%信頼区間:下限=" << odrl << "  上限=" << odru << "\n";
  fout << "  ";
  if (dfh > 0) fout << "調整";
  fout << "オッズ比の検定:χ2乗=" << ca << "  自由度=" << dfa << "  p=" << pr << pind(pr) << "\n";
  if (ftabt) return;            // 全度数表合計
  if (dfh < 1)
  {
    if (opt.res("m") == false) return;
    pr *= (double)nftab;
    if (pr > 1.0) pr = 1.0;
    fout << "  Bonferroni型多重検定:p=" << pr << pind(pr) << "\n";
    ci = ptoz((1.0 - opt.cc()) / (double)nftab) * sqrt(vodr);
    ln_odrl = ln_odr - ci;  ln_odru = ln_odr + ci;
    fout << "    " << opt.cc() * 100.0 << "%同時信頼区間:下限=" << exp(ln_odrl) << "  上限=" << exp(ln_odru) << endl;
    return;
  }
  pr = ctop(ch, dfh);
  fout << "  異質性の検定:χ2乗=" << ch << "  自由度=" << dfh << "  p=" << pr << pind(pr) << endl;
}

void mtestclass::put_emtest(ostream& fout, option& opt, const listftab& ftablst0, int nftab, int nt, bool ftabt) const
// 拡張Mantel検定の結果を出力する
{
  listftab::const_iterator fitr = ftablst0.begin();
  double cmm = 0.0, cmr = 0.0, rd = 0.0, vrd = 0.0;
  int dfm = (int)(fitr->nr()) - 1, dfr = emtest(cmm, cmr, rd, vrd, ftablst0);
  ldbl pr = ctop(cmm, dfm);
  double r2 = cmm / (double)(nt - 1);
  fout << "・拡張Mantel検定\n";
  if (opt.res("r")) fout << "  順位相関比:η=" << sqrt(r2) << "  寄与率:η2乗=" << r2 << "\n";
  double pro = 100.0 / (double)nt;
  double ci = 0.0, rdl = 0.0, rdu = 0.0;
  fout << "  ";
  if (dfr > 0) fout << "調整";
  fout << "順位平均値の差(%)=" << rd << "(" << rd * pro << "%)\n";
  ci = ptoz(1.0 - opt.cc()) * sqrt(vrd);
  rdl = rd - ci;  rdu = rd + ci;
  fout << "    " << opt.cc() * 100.0 << "%信頼区間=" << rdl  << "-" << rdu << "(" << rdl * pro << "%-" << rdu * pro << "%)\n";
  fout << "  共通性の検定:χ2乗=" << cmm << "  自由度=" << dfm << "  p=" << pr << pind(pr) << "\n";
  if (ftabt) return;            // 全度数表合計
  if (dfr < 1)
  {
    if (opt.res("m") == false) return;
    pr *= (ldbl)nftab;
    if (pr > 1.0L) pr = 1.0L;
    fout << "  Bonferroni型多重検定:p=" << pr << pind(pr) << "\n";
    ci = ptoz((1.0 - opt.cc()) / (double)nftab) * sqrt(vrd);
    rdl = rd - ci;  rdu = rd + ci;
    fout << "    " << opt.cc() * 100.0 << "%同時信頼区間=" << rdl  << "-" << rdu << "(" << rdl * pro << "%-" << rdu * pro << "%)\n";
    return;
  }
  pr = ctop(cmr, dfr);
  fout << "  異質性の検定:χ2乗=" << cmr << "  自由度=" << dfr << "  p=" << pr << pind(pr) << endl;
}

void mtestclass::put_gemtest(ostream& fout, option& opt, const listftab& ftablst0, int nftab, int nt, bool ftabt) const
// 一般化拡張Mantel検定の結果を出力する
{
  listftab::const_iterator fitr = ftablst0.begin();
  double cmm = 0.0, cmr = 0.0;
  int dfm = (int)(fitr->nr()) - 1, dfr = gemtest(cmm, cmr, ftablst0);
  ldbl pr = ctop(cmm, dfm);
  double r2 = cmm / (double)(nt - 1);
  fout << "・一般化拡張Mantel検定\n";
  if (opt.res("r")) fout << "  順位相関比:η=" << sqrt(r2) << "  寄与率:η2乗=" << r2 << "\n";
  fout << "  共通性の検定:χ2乗=" << cmm << "  自由度=" << dfm << "  p=" << pr << pind(pr) << "\n";
  if (ftabt) return;            // 全度数表合計
  if (dfr < 1)
  {
    if (opt.res("m") == false) return;
    pr *= (ldbl)nftab;
    if (pr > 1.0L) pr = 1.0L;
    fout << "  Bonferroni型多重検定:p=" << pr << pind(pr) << "\n";
    return;
  }
  pr = ctop(cmr, dfr);
  fout << "  異質性の検定:χ2乗=" << cmr << "  自由度=" << dfr << "  p=" << pr << pind(pr) << endl;
}

void mtestclass::plot_ngp(int fno, option& opt, listidd& idlst, const varclass& var)
// Ngraph用データを出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname(fno)).c_str()) == false) return; // ファイルオープン失敗
  bool delml = false;
  if (opt.ngp("mhrd"))          // 調整リスク差
  {
    fout << "-s1 -dline\n";
    plot_ngpid(fout, idlst);
    plot_rcivec(fout, "M-H調整リスク差", var.nftabok(), var.nt(), var.mhrdvec());
    delml = true;
  }
  if (opt.ngp("mhrr"))          // 調整リスク比
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80\n";
      delml = true;
    }
    plot_ngpid(fout, idlst);
    plot_rcivec(fout, "M-H調整リスク比", var.nftabok(), var.nt(), var.mhrrvec());
  }
  if (opt.ngp("mhor"))          // 調整オッズ比
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80\n";
      delml = true;
    }
    plot_ngpid(fout, idlst);
    plot_rcivec(fout, "M-H調整オッズ比", var.nftabok(), var.nt(), var.mhorvec());
  }
  if (opt.ngp("rr"))            // リスク比
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80\n";
      delml = true;
    }
    plot_ngpid(fout, idlst);
    plot_rcivec(fout, "リスク比", var.nftabok(), var.nt(), var.rrvec());
  }
  if (opt.ngp("or"))            // オッズ比
  {
    if (delml) fout << NGPdelm << "\n";
    else fout << "-s1 -dline -w80\n";
    plot_ngpid(fout, idlst);
    plot_rcivec(fout, "オッズ比", var.nftabok(), var.nt(), var.orvec());
  }
  fout.close();
}

void mtestclass::plot_ngpid(ofstream& fout, listidd& idlst) const
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "%";  puttitle(fout, Title);
  fout << "%\n";
  listidd::iterator iitr = idlst.begin();
  iitr->put_title(fout, formngp);
  iitr->put_stra(fout, formngp);
  for (size_t i = 0; i < iitr->ni(); ++i)
  {
    fout << "%";
    switch (iitr->eitem(i).aval())
    {
    case avalrow : fout << "群項目(縦)  :" << iitr->item(i) << "\n";
      break;
    case avalcol : fout << "順序項目(横):" << iitr->item(i) << "\n";
      break;
    default : iitr->put_item(fout, i);
    }
  }
}

void mtestclass::plot_rcivec(ofstream& fout, const string& rname, int nftab, int nt, const vecrci& rcivec) const
// 個々のリスク比と信頼区間をNgraph用に出力する
{
  int ymax = nftab + 1;
  double xmin = FDmax, xmax = 0.0;
  fout << "%" << rname << "信頼区間(直線)\n";
  for (int i = 0; i < nftab; ++i)
  {
    fout << rcivec[i].rl() << " " << ymax - i << "\n";
    fout << rcivec[i].ru() << " " << ymax - i << "\n";
    fout << NGPdelm2 << "\n";
    if (rcivec[i].rl() < xmin) xmin = rcivec[i].rl();
    if (xmax < rcivec[i].ru()) xmax = rcivec[i].ru();
  }
  double xw = (xmax - xmin) * 0.1, x = 0.0, y = 0.0;
  double yw = (nftab > 8) ? 0.9 : (double)ymax * 0.1;
  fout << NGPdelm << "\n";
  fout << "%" << rname << "点推定(塗りつぶし矩形)\n";
  for (int i = 0; i < nftab; ++i)
  {
    double w = (double)(rcivec[i].n()) / (double)nt;
    x = rcivec[i].r() - xw * 0.5 * w;
    y = (double)(ymax - i) - yw * 0.5 * w;
    fout << x << " " << y << " ";
    x = rcivec[i].r() + xw * 0.5 * w;
    y = (double)(ymax - i) + yw * 0.5 * w;
    fout << x << " " << y << "\n";
  }
  fout << NGPdelm << "\n";
  fout << "%" << rname << "全体信頼区間(実多角形)\n";
  fout << rcivec[nftab + 1].rl() << " " << 1 << "\n";
  y = 1.0 + yw * 0.5;
  fout << rcivec[nftab + 1].r() << " " << y << "\n";
  fout << rcivec[nftab + 1].ru() << " " << 1 << "\n";
  y = 1.0 - yw * 0.5;
  fout << rcivec[nftab + 1].r() << " " << y << endl;
}

int mtestclass::get(const char* c_fname, option& opt, listidd& idlst, listftab& ftablst)
// 度数集計表を読み込む
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nftab = fgetftable(fname, idlst, ftablst);
  if (nftab < 0) throw runtime_error("実行時エラー！");
  if (opt.check() != 0) throw option_error("実行時オプションエラー！");
  return nftab;                 // 読み込んだ度数表数を返す
}

void mtestclass::put(option& opt, listidd& idlst, listftab& ftablst, int fno)
// 集計結果を出力する
{
  varclass var;                 // 計算用変数クラス
  var.cal(opt, idlst, ftablst);
  if (var.nftabok() < 1) throw runtime_error("有効な度数表無し！");
  if (opt.skip() == false && var.nftabok() < (int)(ftablst.size())) throw runtime_error("度数表が不揃い！");
  if (opt.res("res"))
  {
    put_ftable(cout, opt, idlst, ftablst, var);
    if (var.nftabok() > 1 && opt.res("stt") && (opt.res("mt") || opt.res("rrt") || opt.res("lgt") || opt.res("emt") || opt.res("gemt"))) put_overall(cout, opt, ftablst, var);
    cout << "\n" << endl;
  }
  if (opt.ngpname(fno) != "") plot_ngp(fno, opt, idlst, var);
}

int main(int argc, char* argv[]) // メインルーチン
{
  mtestclass mtest;             // メインクラス変数
  option opt(argv[0]);          // オプション
  listidd idlst;                // データ識別情報リスト
  listftab ftablst;             // 度数表リスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    int fno = 0;                // 有効な度数表の数とファイル番号
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok)                // 標準入力から読み込む
        {
          mtest.get(nullptr, opt, idlst, ftablst);
          mtest.put(opt, idlst, ftablst, fno);
          ++fno;
        }
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      {
        mtest.get(argv[i], opt, idlst, ftablst);  fok = false;
        mtest.put(opt, idlst, ftablst, fno);
        ++fno;
      }
    }
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "内容不明なエラー発生！");
    return 255;
  }
  return 0;
}
